from __future__ import annotations

import logging

from copy import deepcopy
from fontTools.designspaceLib import (
    AxisDescriptor,
    AxisLabelDescriptor,
    DesignSpaceDocument,
)
from fontTools.ufoLib import UFOFileStructure
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Tuple
from ufoLib2.objects.font import Font
from ufoLib2.objects.features import Features
from ufonormalizer import normalizeUFO
from vfbLib.ufo.designspace import get_ds_location
from vfbLib.ufo.glyph import VfbToUfoGlyph
from vfbLib.ufo.groups import transform_groups
from vfbLib.ufo.guides import apply_guide_properties, get_master_guides
from vfbLib.ufo.info import VfbToUfoInfo
from vfbLib.ufo.kerning import UfoKerning
from vfbLib.ufo.paths import UfoMasterGlyph
from vfbLib.ufo.tth import TTGlyphHints, transform_stem_rounds
from vfbLib.ufo.typing import (
    TUfoStemPPMsDict,
    TUfoStemsDict,
    TUfoTTZoneDict,
    TUfoTTZonesDict,
)
from vfbLib.ufo.vfb2ufo import TT_GLYPH_LIB_KEY, TT_LIB_KEY

if TYPE_CHECKING:
    from fontTools.designspaceLib import DiscreteAxisDescriptor
    from vfbLib.typing import Anchor, ClassFlagDict, GuidePropertyList
    from vfbLib.ufo.typing import UfoGroups, UfoMMKerning
    from vfbLib.vfb.vfb import Vfb


logger = logging.getLogger(__name__)


class VfbToUfoBuilder:
    def __init__(
        self,
        vfb: Vfb,
        minimal=False,
        base64=False,
        pshints=True,
    ) -> None:
        """
        Serialize the JSON structure to UFO(s)
        """
        self.axes: List[AxisDescriptor | DiscreteAxisDescriptor] = []
        self.axis_count = 0
        self.vfb = vfb
        self.minimal = minimal
        self.encode_data_base64 = base64
        self.include_ps_hints = pshints

        self.features_classes = ""
        self.features_code = ""
        self.groups: UfoGroups = {}
        self.guide_properties: GuidePropertyList = []
        self.info = VfbToUfoInfo()
        self.kerning_class_flags: ClassFlagDict = {}
        self.num_blue_values = 0
        self.num_other_blues = 0
        self.num_family_blues = 0
        self.num_family_other_blues = 0
        self.num_stem_snap_h = 0
        self.num_stem_snap_v = 0
        self.mm_guides = None
        self.mm_kerning: UfoMMKerning = {}
        self.kerning: Dict[Tuple[str, str], int] = {}
        self.lib: Dict[str, Any] = {}
        self.masters: List[str] = []
        self.master_locations: Dict[int, List[float]] = {}
        self.masters_ps_info: List[Dict] = []
        self.primary_instances: List[Dict[str, str | List[float]]] = []
        self.current_glyph: VfbToUfoGlyph | None = None
        self.glyph_masters: Dict[str, VfbToUfoGlyph] = {}
        self.glyphOrder: List[str] = []
        # TT
        self.stem_ppms: TUfoStemPPMsDict = {"ttStemsH": [], "ttStemsV": []}
        self.stems: TUfoStemsDict = {"ttStemsH": [], "ttStemsV": []}
        self.tt_zones: TUfoTTZonesDict = {}
        self.tt_zone_names: List[str] = []
        self.zone_names: Dict[str, List[str]] = {}
        self.build()

    def add_axis_mappings(self, data: List[Tuple[float, float]]) -> None:
        if not self.axis_mappings_count:
            raise ValueError(
                "If axis mappings are present, axis mappings count must be set before "
                "parsing them."
            )
        for i in range(self.axis_count):
            # Get the number of mappings for the current axis
            n = self.axis_mappings_count[i]
            if n > 0:
                # Use only n mappings out of 10 for the current axis
                mappings = data[:n]
                axis = self.axes[i]
                axis.map = [(u, round(d * 1000)) for u, d in mappings]
                if isinstance(axis, AxisDescriptor):
                    # Derive min/max from the mappings
                    # FIXME: Is this info available elsewhere in the VFB?
                    user_coords = [c[0] for c in mappings]
                    axis.minimum = min(user_coords, default=0)
                    axis.default = axis.minimum
                    axis.maximum = max(user_coords, default=1000)
            # Jump to the next 10 axis mappings
            data = data[10:]

    def add_ot_class(self, data: Dict[str, List[str] | str]) -> None:
        if "name" not in data:
            logger.warning(f"Malformed OT class definition, skipping: {data}")
            return

        name = data["name"]
        if not isinstance(name, str):
            raise TypeError(f"OT class name must be a string: {name} ({type(name)})")

        is_kerning = name.startswith("_")

        if name in self.groups:
            logger.warning(f"Duplicate OT class name, skipping: {name}")
            return

        glyphs_list = data["glyphs"]
        if is_kerning:
            # Reorganize glyphs so that the "keyglyph" is first
            glyphs: List[str] = [g.strip() for g in glyphs_list if not g.endswith("'")]
            keyglyphs = [g.strip() for g in glyphs_list if g.endswith("'")]
            keyglyphs = [k.strip("'") for k in keyglyphs]
            if len(keyglyphs) != 1:
                logger.warning(
                    f"Unexpected number of key glyphs in group {name}: {keyglyphs}"
                )
            else:
                glyphs.insert(0, *keyglyphs)

        else:
            glyphs = [g.strip() for g in glyphs_list]
        self.groups[name] = glyphs
        # Also add non-kerning classes to the feature code
        if not is_kerning and not name.startswith("."):
            self.features_classes += f"@{name} = [{' '.join(glyphs)}];\n"

    def assure_tt_lib(self) -> None:
        if TT_LIB_KEY not in self.lib:
            self.lib[TT_LIB_KEY] = {}

    def assure_tt_glyphlib(self) -> None:
        if TT_GLYPH_LIB_KEY not in self.lib:
            self.lib[TT_GLYPH_LIB_KEY] = {}

    def build_mm_glyph(self, data: Dict[str, Any]) -> None:
        g = self.current_glyph = VfbToUfoGlyph()
        g.lib = {}
        g.name = data["name"]
        g.unicodes = []

        # MM Stuff, need to extract later
        if not self.minimal and "guides" in data:
            g.mm_guides = data["guides"]

        if self.include_ps_hints and "hints" in data:
            g.mm_hints = data["hints"]
            if "hintmasks" in data["hints"]:
                g.hintmasks = data["hints"]["hintmasks"]
                del data["hints"]["hintmasks"]

        if "kerning" in data:
            kerning: Dict[int, List[int]] = data["kerning"]
            for Rid, values in kerning.items():
                assert g.name is not None
                self.mm_kerning[(g.name, Rid)] = values

        g.mm_metrics = data["metrics"]  # width and height
        g.mm_nodes = data["nodes"]

        if "components" in data:
            g.mm_components = data["components"]

        # TrueType hinting, needs to come after mm_nodes, because it needs
        # access to the point indices.
        if "tth" in data:
            tth = TTGlyphHints(g, data["tth"], self.zone_names, self.stems)
            g.tth_commands = tth.commands
            del tth

    def set_glyph_background(self, data: Dict[str, Any]) -> None:
        assert self.current_glyph is not None
        self.current_glyph.lib["com.fontlab.v5.background"] = data

    def set_tt_stem_ppms(self, data: Dict[str, List[Dict[str, Any]]]) -> None:
        for d in ("ttStemsH", "ttStemsV"):
            direction_stems = data[d]
            for ds in direction_stems:
                rounds = transform_stem_rounds(ds["round"], d)
                stem = {
                    "index": ds["stem"],
                    "round": rounds,
                }
                self.stem_ppms[d].append(stem)

    def set_tt_stems(self, data) -> None:
        for d in ("ttStemsH", "ttStemsV"):
            direction_stems = data[d]
            for i, ds in enumerate(direction_stems):
                r = ds["round"]
                rk = list(r.keys())[0]
                rv = str(r[rk])
                stem_ppms = self.stem_ppms[d][i]
                assert i == stem_ppms["index"]
                stem = {
                    "horizontal": d == "ttStemsV",
                    "width": ds["value"],
                    "name": ds["name"],
                    "round": stem_ppms["round"],
                }
                if rv in stem["round"]:
                    logger.error(
                        f"Error in stem rounding settings for {d}, duplicate ppm {rv}."
                    )
                stem["round"][rv] = int(rk)
                self.stems[d].append(stem)

    def set_tt_zones(self, data: Dict[str, List]) -> None:
        self.zone_names = {
            "ttZonesT": [],
            "ttZonesB": [],
        }
        for d in ("ttZonesB", "ttZonesT"):
            direction_zones = data[d]
            for dz in direction_zones:
                zone = TUfoTTZoneDict(
                    position=dz["position"],
                    top=d == "ttZonesT",
                    width=dz["value"],
                )
                name = dz["name"]
                if name in self.tt_zones:
                    logger.warning(f"Duplicate zone name: {name}, overwriting.")
                self.tt_zones[name] = zone
                self.tt_zone_names.append(name)  # for deltas
                self.zone_names[d].append(name)  # for AlignTop/AlignBottom

    def set_tt_pixel_snap(self, data) -> None:
        self.assure_tt_lib()
        if data != 68:
            # Omit default value
            self.lib[TT_LIB_KEY]["stemsnaplimit"] = data

    def set_tt_zone_stop(self, data) -> None:
        self.assure_tt_lib()
        if data != 48:
            # Omit default value
            self.lib[TT_LIB_KEY]["zoneppm"] = data

    def set_tt_code_stop(self, data) -> None:
        self.assure_tt_lib()
        self.lib[TT_LIB_KEY]["codeppm"] = data

    def set_tt_zone_deltas(self, data) -> None:
        for zone_index, deltas in data.items():
            zone_name = self.tt_zone_names[int(zone_index)]
            self.tt_zones[zone_name]["delta"] = {str(k): v for k, v in deltas.items()}

    def build_tt_stems_lib(self) -> None:
        lib = {}
        for d in ("ttStemsH", "ttStemsV"):
            i = 0
            direction_stems = self.stems[d]
            for stem in direction_stems:
                name = stem["name"]
                del stem["name"]
                if name == "":
                    dname = "v" if d == "ttStemsH" else "h"
                    name = "%s%02i" % (dname, i)
                    i += 1
                if name in lib:
                    oldname = name
                    name = "%s#%02i" % (oldname, i)
                    logger.warning(
                        f"Duplicate TrueType stem name '{oldname}', renamed to '{name}'"
                        ". Preferably make stem names unique in VFB."
                    )
                lib[name] = stem
        if lib:
            self.lib[TT_LIB_KEY]["stems"] = lib

    def build_tt_zones_lib(self) -> None:
        self.assure_tt_lib()
        if self.tt_zones:
            self.lib[TT_LIB_KEY]["zones"] = self.tt_zones

    def build(self) -> None:  # noqa: C901
        # Non-MM data
        for e in self.vfb.entries:
            name = e.key
            if name is None:
                raise TypeError
            data = e.decompiled
            if data is None:
                continue

            # Font Info
            attr = self.info.mapping.get(name, None)
            if attr is not None:
                self.info.set_attr(attr, data)
                continue

            if name == "header":
                pass
            elif name == "sgn":  # 1024
                self.info.familyName = data
                self.info.openTypeNamePreferredFamilyName = data
            elif name == "Monospaced":  # 1034
                self.info.set_fixed_pitch(data)
            elif name == "version full":
                pass
            elif name == "weight":  # 1048
                self.info.set_weight_class(data)
            elif name == "Style Name":  # 1127
                self.info.postscriptWeightName = data  # FIXME: Does not contain Italic
            elif name == "Type 1 XUIDs":  # 1133
                pass
            elif name == "tsn":  # 1137
                self.info.set_style_name(data)
            elif name == "Name Records":  # 1138
                self.info.set_name_records(data)
            elif name == "Glyph Unicode":  # 1250
                assert self.current_glyph is not None
                self.current_glyph.unicodes.extend(data)
            elif name == "Glyph Unicode Non-BMP":  # 1253
                assert self.current_glyph is not None
                self.current_glyph.unicodes.extend(data)
            elif name == "Primary Instances":  # 1254
                self.primary_instances = data
            elif name == "TrueType Zones":  # 1255
                self.set_tt_zones(data)
            elif name == "TrueType Info":  # 1264
                self.info.set_tt_info(data)
            elif name == "Gasp Ranges":  # 1265
                self.info.set_tt_gasp(data)
            elif name == "Selection":  # 1267
                self.info.set_selection(data)
            elif name == "TrueType Stem PPEMs":  # 1268
                self.set_tt_stem_ppms(data)
            elif name == "TrueType Stems":  # 1269
                self.set_tt_stems(data)
            elif name == "Pixel Snap":  # 1272
                self.set_tt_pixel_snap(data)
            elif name == "TrueType Zone Deltas":  # 1273
                self.set_tt_zone_deltas(data)
            elif name == "Zone Stop PPEM":  # 1274
                self.set_tt_zone_stop(data)
            elif name == "Code Stop PPEM":  # 1275
                self.set_tt_code_stop(data)
            elif name == "openTypeFeatures":  # 1276
                self.features_code = "\n".join(data)
            elif name == "OpenType Class":  # 1277
                self.add_ot_class(data)
            elif name == "Global Guides":  # 1294
                self.mm_guides = data
            elif name == "Global Guide Properties":  # 1296
                self.guide_properties = data
            elif name == "Encoding":  # 1500
                pass
            elif name == "Master Count":  # 1503
                self.master_count = data
            elif name == "Master Name":  # 1504
                self.masters.append(data)
            elif name == "Master Location":  # 1505
                index, locations = data
                self.master_locations[index] = locations
            elif name == "Axis Count":  # 1513
                self.axis_count = data
            elif name == "Axis Name":  # 1514
                tags = {
                    "Weight": "wght",
                    "Width": "wdth",
                    "Optical Size": "opsz",
                    "Serif": "SERF",
                }
                self.axes.append(
                    AxisDescriptor(tag=tags.get(data, data.upper()[:4]), name=data)
                )
            elif name == "Axis Mappings Count":  # 1515
                self.axis_mappings_count = data
            elif name == "Axis Mappings":  # 1516
                self.add_axis_mappings(data)
            elif name == "Anisotropic Interpolation Mappings":  # 1523
                # TODO: Can we properly output this to designspace?
                for axis in data:
                    for a, b in axis:
                        if a != b:
                            maps = {
                                self.axes[i].name: data[i]
                                for i in range(self.axis_count)
                            }
                            logger.warning(
                                "WARNING: Designspace output of anisotropic "
                                "interpolation settings is not yet supported. You must "
                                "set it up manually."
                            )
                            logger.warning(f"         Mappings: {maps}")
                            break
            elif name == "Blue Values Count":  # 1530
                self.num_blue_values = data
            elif name == "Other Blues Count":  # 1531
                self.num_other_blues = data
            elif name == "Family Blues Count":  # 1532
                self.num_family_blues = data
            elif name == "Family Other Blues Count":  # 1533
                self.num_family_other_blues = data
            elif name == "StemSnapH Count":  # 1534
                self.num_stem_snap_h = data
            elif name == "StemSnapV Count":  # 1535
                self.num_stem_snap_v = data
            elif name == "PostScript Info":  # 1536
                self.masters_ps_info.append(data)
            elif name == "Glyph":  # 2001
                if self.current_glyph is not None:
                    name = self.current_glyph.name
                    if name in self.glyph_masters:
                        for i in range(255):
                            n = f"{name}#{i}"
                            if n not in self.glyph_masters:
                                self.glyph_masters[n] = self.current_glyph
                                self.glyphOrder.append(n)
                                logger.warning(
                                    f"Duplicate glyph, renamed from {name} to {n}."
                                )
                                break
                    else:
                        if name is None:
                            logger.error("Glyph without name.")
                        else:
                            self.glyph_masters[name] = self.current_glyph
                            self.glyphOrder.append(name)
                self.build_mm_glyph(data)
            elif name == "Background Bitmap":  # 2007
                self.set_glyph_background(data)
            elif name == "Links":  # 2008
                assert self.current_glyph is not None
                self.current_glyph.links = data
            elif name == "Mask":  # 2009
                pass
            elif name == "2010":
                pass
            elif name == "2011":
                pass
            elif name == "Mark Color":  # 2012
                assert self.current_glyph is not None
                self.current_glyph.set_mark(data)
            elif name == "Glyph User Data":  # 2015
                assert self.current_glyph is not None
                self.current_glyph.lib["com.fontlab.v5.userData"] = data
            elif name == "Font User Data":  # 2016
                self.lib["com.fontlab.v5.userData"] = data
            elif name == "Glyph Note":  # 2017
                assert self.current_glyph is not None
                self.current_glyph.note = data
            elif name == "Glyph GDEF Data":  # 2018
                if "anchors" in data:
                    assert self.current_glyph is not None
                    self.current_glyph.anchors = []
                    for anchor in data["anchors"]:
                        a: Anchor = {
                            "x": anchor["x"],
                            "y": anchor["y"],
                        }
                        if "name" in anchor:
                            a["name"] = anchor["name"]
                        self.current_glyph.anchors.append(a)
            elif name == "Glyph Anchors Supplemental":  # 2020
                pass
            elif name == "Unicode Ranges":  # 2021
                self.info.set_unicode_ranges(data)
            elif name == "2023":
                pass
            elif name == "OpenType Metrics Class Flags":  # 2024
                self.lib["com.fontlab.v5.metricsClassFlags"] = data
            elif name == "OpenType Kerning Class Flags":  # 2026
                self.kerning_class_flags = data
            elif name == "Glyph Origin":  # 2027
                pass
            elif name == "Glyph Anchors MM":  # 2029
                assert self.current_glyph is not None
                self.current_glyph.mm_anchors = data
            elif name == "Glyph Guide Properties":  # 2031
                assert self.current_glyph is not None
                self.current_glyph.guide_properties = data
            else:
                logger.info(f"Unhandled key: {name}")

        if self.current_glyph is not None:
            assert self.current_glyph.name is not None
            self.glyph_masters[self.current_glyph.name] = self.current_glyph
            self.glyphOrder.append(self.current_glyph.name)
        self.lib["public.glyphOrder"] = self.glyphOrder
        self.assure_tt_lib()
        self.build_tt_stems_lib()
        self.build_tt_zones_lib()
        if not self.lib[TT_LIB_KEY]:
            del self.lib[TT_LIB_KEY]
        self.info.fix_underline_position()

    def get_master_info(self, master_index: int = 0) -> VfbToUfoInfo:
        master_info = deepcopy(self.info)
        # Update the info with master-specific values
        properties = [
            ("ascender", "ascender"),
            ("descender", "descender"),
            ("x_height", "xHeight"),
            ("cap_height", "capHeight"),
        ]
        if self.include_ps_hints:
            properties.extend(
                [
                    # ("force_bold", "postscriptForceBold"),
                    ("blue_scale", "postscriptBlueScale"),
                    ("blue_shift", "postscriptBlueShift"),
                    ("blue_fuzz", "postscriptBlueFuzz"),
                ]
            )

        for key, attr in properties:
            value = self.masters_ps_info[master_index].get(key, None)
            if value is not None:
                setattr(master_info, attr, value)

        if self.include_ps_hints:
            # Set "force bold"
            value = self.masters_ps_info[master_index].get("force_bold", None)
            if value is not None:
                master_info.postscriptForceBold = bool(value)

            # Set attributes that must be clipped at a certain index
            for key, attr in (
                ("blue_values", "postscriptBlueValues"),
                ("other_blues", "postscriptOtherBlues"),
                ("family_blues", "postscriptFamilyBlues"),
                ("family_other_blues", "postscriptFamilyOtherBlues"),
                ("stem_snap_h", "postscriptStemSnapH"),
                ("stem_snap_v", "postscriptStemSnapV"),
            ):
                # Get the number of values from the corresponding attribute
                num_values = getattr(self, f"num_{key}")
                if num_values == 0:
                    continue

                value = self.masters_ps_info[master_index].get(key, None)
                if value is not None:
                    setattr(master_info, attr, value[:num_values])

        # Guides
        if not self.minimal and self.mm_guides is not None:
            guides = get_master_guides(self.mm_guides, master_index)
            apply_guide_properties(guides, self.guide_properties)

            if guides:
                master_info.guidelines = guides

        return master_info

    def get_ufo_master(self, index: int, silent=False) -> Font:
        if not silent:
            print(f"Processing font: {self.info.ui_name.strip()}, master {index}")

        # Build the master-specific data that can be passed when instantiating the UFO
        master_kerning = self.ufo_kerning.get_master_kerning(master_index=index)
        master_info = self.get_master_info(master_index=index)

        # Pass as much data right to the UFO
        ufo = Font(
            features=self.ufo_features.text,
            groups=deepcopy(self.ufo_groups),
            info=master_info,
            kerning=master_kerning,
            lib=self.lib,
        )

        # Add the glyphs
        for name, mm_glyph in self.glyph_masters.items():
            logger.debug(f"    {name}, {type(name)}, {mm_glyph}")
            # TODO: Any way to use the ufoLib2 Glyph directly?
            master_glyph = UfoMasterGlyph(mm_glyph, self.glyphOrder, index)
            master_glyph.build(
                self.minimal, self.include_ps_hints, self.encode_data_base64
            )
            ufo_glyph = ufo.newGlyph(name)
            pen = ufo_glyph.getPointPen()
            master_glyph.drawPoints(pen)
            # FIXME: Anchors work, but typing wrong
            ufo_glyph.anchors = master_glyph.anchors
            ufo_glyph.guidelines = master_glyph.guidelines
            ufo_glyph.height = master_glyph.height
            ufo_glyph.lib = master_glyph.lib
            ufo_glyph.width = master_glyph.width
            ufo_glyph.unicodes = master_glyph.unicodes

        return ufo

    def get_ufo_masters(self, silent=False) -> List[Font]:
        # Prepare data shared by the master UFOs
        self.ufo_groups = transform_groups(
            self.groups,
            self.kerning_class_flags,
            self.glyphOrder,
            self.minimal,
        )
        self.ufo_kerning = UfoKerning(self.glyphOrder, self.ufo_groups, self.mm_kerning)
        self.ufo_groups = self.ufo_kerning.groups
        self.ufo_features = Features(self.features_code)
        if self.features_classes:
            self.ufo_features.text = self.features_classes + "\n\n" + self.features_code

        ufo_masters = []
        for i in range(len(self.masters)):
            ufo_masters.append(self.get_ufo_master(i, silent))
        return ufo_masters

    def get_ufos_designspace(
        self, out_path: Path, silent=False
    ) -> Tuple[List[Font], DesignSpaceDocument | None]:
        """
        Build UFOs and a DesignSpaceDocument from the VFB contents in memory and return
        them. The DesignSpaceDocument is only returned for VFBs containing more than one
        master. In other cases, the second element of the returned tuple is None.
        """
        ufo_masters = self.get_ufo_masters(silent)
        ds: DesignSpaceDocument | None = None
        if self.axes:
            ds = self.get_designspace(out_path)
        return ufo_masters, ds

    def write(self, out_path: Path, overwrite=False, silent=False, ufoz=False) -> None:
        """
        Write a the VFB contents to master UFOs and a designspace file. The designspace
        file is only written if the VFB contains more than one master.
        """
        # Build UFOs and DesignSpace
        ufos, ds = self.get_ufos_designspace(out_path, silent)

        # Write the master UFOs
        strct = UFOFileStructure.ZIP if ufoz else None
        for index, ufo in enumerate(ufos):
            master_path = self.get_master_path(out_path, index)
            ufo.save(master_path, structure=strct, overwrite=overwrite)
            if not ufoz:
                normalizeUFO(
                    ufoPath=master_path, onlyModified=False, writeModTimes=False
                )

        # Write the Designspace
        if ds:
            ds_path = out_path.with_suffix(".designspace")
            if ds_path.exists():
                if not overwrite:
                    raise FileExistsError(str(ds_path))
            if not silent:
                print(f"Writing designspace: {ds_path}")
            ds.write(str(ds_path))

    def get_designspace(self, out_path: Path) -> DesignSpaceDocument:
        """
        Build and return a DesignSpaceDocument. The out_path argument will be used to
        construct the UFO paths assigned to the source descriptors.
        """
        ds = DesignSpaceDocument()
        ds.axes = self.axes

        # Add labels
        if self.axis_count == 1:
            # Only support the simple case here
            for p in self.primary_instances:
                self.axes[0].axisLabels.append(
                    AxisLabelDescriptor(name=p["name"], userValue=p["values"][0])
                )
        else:
            logger.warning(
                "WARNING: Designspace output of AxisLabelDescriptors is only supported "
                "for VFBs with 1 axis. Please add labels manually."
            )

        # Add sources
        for i in range(self.master_count):
            ds.addSourceDescriptor(
                # designLocation=,
                familyName=self.info.ds_family_name,
                filename=f"{self.get_master_path(out_path.with_suffix('.ufo'), i)}",
                location=get_ds_location(self.axes, self.master_locations[i + 1]),
                name=self.masters[i],
                styleName=f"Master {i}",
            )

        # Add instances
        for i in range(len(self.primary_instances)):
            p = self.primary_instances[i]
            loc = p["values"]
            if not isinstance(loc, list):
                raise TypeError

            style_name = p["name"]
            # FIXME: Support anisotropic interpolation
            ds.addInstanceDescriptor(
                # designLocation=get_ds_design_location(self.axes, loc),
                familyName=self.info.ds_family_name,
                filename=f"instance_ufo/{self.info.ds_family_name}-{style_name}.ufo",
                # lib=,
                # locationLabel="",
                name=f"instance_{i}",
                # postScriptFontName="",
                # FIXME:
                styleMapFamilyName=f"{self.info.ds_family_name} {style_name}".strip(),
                styleMapStyleName=self.info.styleMapStyleName,
                styleName=style_name,
                userLocation=get_ds_location(self.axes, loc, 1),
            )

        return ds

    @classmethod
    def get_master_path(cls, out_path: Path, master_index: int) -> Path:
        if master_index > 0:
            try:
                return out_path.with_stem(f"{out_path.stem}-{master_index}")
            except AttributeError:
                # Python 3.8 doesn't have `Path.with_stem`
                return out_path.with_name(
                    f"{out_path.stem}-{master_index}{out_path.suffix}"
                )
        return out_path



from __future__ import annotations

import logging

from vfbLib.parsers import BaseParser
from vfbLib.parsers.value import read_encoded_value


logger = logging.getLogger(__name__)


class CustomCmapParser(BaseParser):
    """
    A parser that reads the custom CMAP settings.
    """

    @classmethod
    def _parse(cls):
        count = read_encoded_value(cls.stream)  # number of cmap records, not values
        values = {"values": []}
        for _ in range(count):
            cmap = {
                "language_id": read_encoded_value(cls.stream),
                "platform_id": read_encoded_value(cls.stream),
                "encoding_id": read_encoded_value(cls.stream),
                "format": read_encoded_value(cls.stream),
                "option": read_encoded_value(cls.stream),
                "records": [],
            }
            num_encodings = read_encoded_value(cls.stream)
            if num_encodings > 0:
                cmap["records"] = [cls.read_uint8() for _ in range(num_encodings)]
            values["values"].append(cmap)
        return values



from __future__ import annotations

import logging

from fontTools.misc.textTools import hexStr  # , num2binary
from fontTools.ttLib.tables.ttProgram import Program
from io import BytesIO
from struct import unpack
from typing import Any, Dict, List, Literal
from vfbLib.parsers import BaseParser, read_encoded_value
from vfbLib.parsers.guides import parse_guides
from vfbLib.truetype import TT_COMMANDS
from vfbLib.typing import (
    Anchor,
    Component,
    GdefDict,
    GlyphData,
    Hint,
    HintDict,
    Instruction,
    LinkDict,
    MaskData,
    MMAnchor,
    MMNode,
    Point,
)


logger = logging.getLogger(__name__)

cmd_name = {
    0: "move",
    1: "line",
    3: "curve",
    4: "qcurve",
}

MOVE = 0
LINE = 1
CURVE = 3
QCURVE = 4


def read_absolute_point(
    points: List[List[Any]],
    stream,
    num_masters: int,
    x_masters: List[int],
    y_masters: List[int],
):
    # Read coordinates for all masters from stream, add the points to a list
    # of points per master. Keep track of the relative coordinates reference
    # value in x_masters and y_masters.
    for m in range(num_masters):
        # End point
        xrel = read_encoded_value(stream)
        yrel = read_encoded_value(stream)
        x_masters[m] += xrel
        y_masters[m] += yrel
        points[m].append((x_masters[m], y_masters[m]))


class GlyphAnchorsParser(BaseParser):
    @classmethod
    def _parse(cls) -> List[MMAnchor]:
        stream = cls.stream
        anchors = []
        num_anchors = read_encoded_value(stream)
        num_masters = read_encoded_value(stream)
        for _ in range(num_anchors):
            anchor = MMAnchor(x=[], y=[])
            for _ in range(num_masters):
                anchor["x"].append(read_encoded_value(stream))
                anchor["y"].append(read_encoded_value(stream))
            anchors.append(anchor)
        return anchors


class GlyphAnchorsSuppParser(BaseParser):
    @classmethod
    def _parse(cls) -> List:
        stream = cls.stream
        anchors = []
        num_anchors = read_encoded_value(stream)
        for _ in range(num_anchors):
            a = read_encoded_value(stream)
            b = read_encoded_value(stream)
            anchors.append([a, b])
        return anchors


class GlyphGDEFParser(BaseParser):
    @classmethod
    def _parse(cls) -> GdefDict:
        stream = cls.stream
        gdef: GdefDict = {}
        class_names = {
            0: "unassigned",
            1: "base",
            2: "ligature",
            3: "mark",
            4: "component",
        }
        glyph_class = read_encoded_value(stream)
        glyph_class_name = class_names.get(glyph_class, "unassigned")
        if glyph_class_name != "unassigned":
            gdef["glyph_class"] = glyph_class_name

        num_anchors = read_encoded_value(stream)
        anchors = []
        for _ in range(num_anchors):
            anchor_name_length = read_encoded_value(stream)
            name = None
            if anchor_name_length > 0:
                name = stream.read(anchor_name_length).decode(cls.encoding)

            x = read_encoded_value(stream)
            x1 = read_encoded_value(stream)
            y = read_encoded_value(stream)
            y1 = read_encoded_value(stream)
            anchor = Anchor(x=x, x1=x1, y=y, y1=y1)
            if name:
                anchor["name"] = name
            anchors.append(anchor)
        if anchors:
            gdef["anchors"] = anchors

        num_carets = read_encoded_value(stream)
        carets = []
        for _ in range(num_carets):
            pos = read_encoded_value(stream)
            xxx = read_encoded_value(stream)
            carets.append((pos, xxx))
        if carets:
            gdef["carets"] = carets

        try:
            num_values = read_encoded_value(stream)
            values = [read_encoded_value(stream) for _ in range(num_values)]
            if values:
                gdef["unknown"] = values
        except EOFError:
            pass

        assert stream.read() == b""
        return gdef


class GlyphOriginParser(BaseParser):
    @classmethod
    def _parse(cls) -> Dict[str, Any]:
        stream = cls.stream
        x = int.from_bytes(stream.read(2), signed=True, byteorder="little")
        y = int.from_bytes(stream.read(2), signed=True, byteorder="little")
        return {"x": x, "y": y}


class GlyphParser(BaseParser):
    @classmethod
    def parse_guides(cls, stream: BytesIO, glyphdata: GlyphData, num_masters=1) -> None:
        # Guidelines
        guides = parse_guides(stream, num_masters)
        if guides:
            glyphdata["guides"] = guides

    @classmethod
    def parse_binary(cls, stream: BytesIO, glyphdata: GlyphData) -> None:
        # Imported binary TrueType data
        imported: Dict[str, Any] = {}
        while True:
            key = cls.read_uint8()

            if key == 0x28:
                break

            elif key == 0x29:
                # Metrics
                imported["width"] = read_encoded_value(cls.stream)
                imported["lsb"] = read_encoded_value(cls.stream)
                imported["unknown1"] = read_encoded_value(cls.stream)
                imported["unknown2"] = read_encoded_value(cls.stream)
                imported["unknown3"] = read_encoded_value(cls.stream)
                imported["bbox"] = [read_encoded_value(cls.stream) for _ in range(4)]

            elif key == 0x2A:
                # Outlines
                num_contours = read_encoded_value(cls.stream)
                imported["endpoints"] = [
                    read_encoded_value(cls.stream) for _ in range(num_contours)
                ]
                num_nodes = read_encoded_value(cls.stream)
                nodes = []
                # logger.debug(f"Parsing {num_nodes} nodes...")
                x = 0
                y = 0
                for i in range(num_nodes):
                    x += read_encoded_value(cls.stream)
                    y += read_encoded_value(cls.stream)
                    byte = cls.read_uint8(stream)
                    flags = byte >> 4
                    cmd = byte & 0x0F
                    node = (hex(cmd), hex(flags), x, y)
                    # logger.debug(f"    {i}: {node}")
                    nodes.append(node)
                if nodes:
                    imported["nodes"] = nodes

            elif key == 0x2B:
                # Instructions
                num_instructions = read_encoded_value(cls.stream)
                instructions = cls.stream.read(num_instructions)
                p = Program()
                p.fromBytecode(instructions)
                imported["instructions"] = p.getAssembly()

            elif key == 0x2C:
                # Probably HDMX data
                num = read_encoded_value(cls.stream)
                imported["hdmx"] = [cls.read_uint8() for _ in range(num)]

            else:
                logger.error(imported)
                logger.error(f"Unknown key in imported binary glyph data: {hex(key)}")
                raise ValueError

        glyphdata["imported"] = imported

    @classmethod
    def parse_components(
        cls, stream: BytesIO, glyphdata: GlyphData, num_masters=1
    ) -> None:
        components = []
        num = read_encoded_value(stream)
        for i in range(num):
            gid = read_encoded_value(stream)
            c = Component(gid=gid, offsetX=[], offsetY=[], scaleX=[], scaleY=[])
            for _ in range(num_masters):
                x = read_encoded_value(stream)
                y = read_encoded_value(stream)
                scaleX, scaleY = unpack("dd", stream.read(16))
                c["offsetX"].append(x)
                c["offsetY"].append(y)
                c["scaleX"].append(scaleX)
                c["scaleY"].append(scaleY)
            components.append(c)
        glyphdata["components"] = components

    @classmethod
    def parse_hints(cls, stream: BytesIO, glyphdata: GlyphData, num_masters=1) -> None:
        hints = HintDict(v=[], h=[])
        for d in ("h", "v"):
            num_hints = read_encoded_value(stream)
            for _ in range(num_hints):
                master_hints = []
                for _ in range(num_masters):
                    pos = read_encoded_value(stream)
                    width = read_encoded_value(stream)
                    master_hints.append(Hint(pos=pos, width=width))
                hints[d].append(master_hints)

        num_hintmasks = read_encoded_value(stream)
        if num_hintmasks > 0:
            hintmasks: List[Dict[str, int]] = []
            for i in range(num_hintmasks):
                k = cls.read_uint8(stream)
                val = read_encoded_value(stream)
                mask = {}
                if k == 0x01:
                    # hintmask for hstem
                    if "h" in mask:
                        raise KeyError
                    mask["h"] = val  # num2binary(val, bits=8)
                elif k == 0x02:
                    # hintmask for vstem
                    if "v" in mask:
                        raise KeyError
                    mask["v"] = val  # num2binary(val, bits=8)
                elif k == 0xFF:
                    # Replacement point
                    # FIXME: This seems to be the node index of the replacement
                    # point. But sometimes it is negative, why?
                    mask["r"] = val
                hintmasks.append(mask)
            if hintmasks:
                hints["hintmasks"] = hintmasks

        if hints["v"] or hints["h"] or "hintmasks" in hints:
            glyphdata["hints"] = hints

    @classmethod
    def parse_instructions(cls, stream: BytesIO, glyphdata: GlyphData) -> None:
        # Number of bytes for instructions that follow;
        # we don't use it
        _ = read_encoded_value(stream)
        num_commands = read_encoded_value(stream)
        commands: List[Instruction] = []
        for i in range(num_commands):
            cmd = cls.read_uint8(stream)
            params = [
                read_encoded_value(stream)
                for _ in range(len(TT_COMMANDS[cmd]["params"]))
            ]
            commands.append(
                Instruction(
                    cmd=TT_COMMANDS[cmd]["name"],
                    params=dict(zip(TT_COMMANDS[cmd]["params"], params)),
                )
            )
        # Instructions are ended by 3 * 0!?
        for _ in range(3):
            read_encoded_value(stream)

        if commands:
            glyphdata["tth"] = commands

    @classmethod
    def parse_metrics(
        cls, stream: BytesIO, glyphdata: GlyphData, num_masters=1
    ) -> None:
        metrics: List[Point] = []
        for _ in range(num_masters):
            master_metrics = (
                read_encoded_value(stream),
                read_encoded_value(stream),
            )
            metrics.append(master_metrics)
        glyphdata["metrics"] = metrics

    @classmethod
    def parse_outlines(cls, stream: BytesIO, glyphdata: GlyphData) -> int:
        # Nodes
        num_masters = read_encoded_value(stream)
        glyphdata["num_masters"] = num_masters

        # 2 x the number of values to be read after num_nodes, the reason is unclear.
        _ = read_encoded_value(stream)
        # glyphdata["num_node_values"] = num_node_values

        num_nodes = read_encoded_value(stream)
        segments: List[MMNode] = []
        x = [0 for _ in range(num_masters)]
        y = [0 for _ in range(num_masters)]

        for _ in range(num_nodes):
            byte = cls.read_uint8(stream)
            flags = byte >> 4
            cmd = byte & 0x0F

            segment_type = cmd_name[cmd]
            # logger.debug(f"    {i}: {segment_type}, flags: {flags}")

            # End point
            points: List[List[Point]] = [[] for _ in range(num_masters)]
            read_absolute_point(points, stream, num_masters, x, y)

            if cmd == CURVE:
                # First control point
                read_absolute_point(points, stream, num_masters, x, y)
                # Second control point
                read_absolute_point(points, stream, num_masters, x, y)

            t: Literal["move", "curve", "line", "qcurve"] = segment_type
            segment = MMNode(type=t, flags=flags, points=points)
            segments.append(segment)
        glyphdata["nodes"] = segments
        return num_masters

    @classmethod
    def parse_kerning(
        cls, stream: BytesIO, glyphdata: GlyphData, num_masters=1
    ) -> None:
        num = read_encoded_value(stream)
        kerning = {}
        for _ in range(num):
            # Glyph index of right kerning partner
            gid = read_encoded_value(stream)
            values = []
            for _ in range(num_masters):
                # List of values, one value per master
                values.append(read_encoded_value(stream))
            kerning[gid] = values
        glyphdata["kerning"] = kerning

    @classmethod
    def _parse(cls) -> Dict[str, Any]:
        """
        01090701
        01  92[7]2e 6e 6f 74 64 65 66 # Glyph name
        08  8c bb[48] 93[8]
            00 ab 8b    [  32,    0] move
            01 8b f9a0  [   0,  780] line 32,780
            01 f7fc 8b  [ 360,    0] line 392,780
            01 8b fda0  [   0, -780] line 392, 0
            00 fbcc bb  [-312,   48] move  80, 48
            01 f79c 8b  [ 264,    0] line
            01 8b f940  [   0,  684] line
            01 fb9c 8b  [-264,    0] line
        02  f83c 8b  [ 424,    0]
        03  8b 8b 8b
        04  8b 8b
        0a  a1[22] 8f[4] # TT hinting
            02 8b 8b
            04 8b 8c 89 8a
            04 8b 8f 89 8a
            04 8c 92 89 8a
            8b 8b 8b
        0f
        """
        s = cls.stream
        glyphdata = GlyphData()
        start = unpack("<4B", s.read(4))
        if start != (1, 9, 7, 1):
            logger.warning(
                f"Unexpected glyph constant: {start}, please notify developer"
            )
        # glyphdata["constants"] = start
        num_masters = 1
        while True:
            # Read a value to decide what kind of information follows
            v = cls.read_uint8(s)

            if v == 0x01:
                # Glyph name
                glyph_name_length = read_encoded_value(s)
                glyph_name = s.read(glyph_name_length)
                glyphdata["name"] = glyph_name.decode(cls.encoding)
                logger.debug(f"Glyph: {glyphdata['name']}")

            elif v == 0x02:
                # Metrics
                cls.parse_metrics(s, glyphdata, num_masters)

            elif v == 0x03:
                # PS Hints
                cls.parse_hints(s, glyphdata, num_masters)

            elif v == 0x04:
                # Guides
                cls.parse_guides(s, glyphdata, num_masters)

            elif v == 0x05:
                # Components
                cls.parse_components(s, glyphdata, num_masters)

            elif v == 0x06:
                # Kerning
                cls.parse_kerning(s, glyphdata, num_masters)

            elif v == 0x08:
                # Outlines
                num_masters = cls.parse_outlines(s, glyphdata)

            elif v == 0x09:
                # Imported binary TrueType data
                cls.parse_binary(s, glyphdata)

            elif v == 0x0A:
                # TrueType instructions
                cls.parse_instructions(s, glyphdata)

            elif v == 0x0F:
                logger.debug("Glyph done.")
                break

            else:
                logger.error(f"Unhandled info field: {hex(v)}")
                logger.error(hexStr(s.read()))
                raise ValueError

        return dict(glyphdata)


class GlyphUnicodeParser(BaseParser):
    @classmethod
    def _parse(cls) -> List:
        unicodes = []
        for _ in range(cls.stream.getbuffer().nbytes // 2):
            u = cls.read_uint16(cls.stream)
            unicodes.append(u)
        return unicodes


class GlyphUnicodeSuppParser(BaseParser):
    @classmethod
    def _parse(cls) -> List:
        unicodes = []
        for _ in range(cls.stream.getbuffer().nbytes // 4):
            u = cls.read_uint32(cls.stream)
            unicodes.append(u)
        return unicodes


class LinkParser(BaseParser):
    @classmethod
    def _parse(cls) -> Dict[str, Any]:
        s = cls.stream
        links = LinkDict(x=[], y=[])
        for i in range(2):
            num = read_encoded_value(s)
            for _ in range(num):
                src = read_encoded_value(s)
                tgt = read_encoded_value(s)
                links["yx"[i]].append([src, tgt])
        return dict(links)


class MaskParser(GlyphParser):
    @classmethod
    def _parse(cls) -> Dict[str, Any]:
        s = cls.stream
        num = read_encoded_value(s)
        maskdata = MaskData(num=num)
        for i in range(num):
            maskdata[f"reserved{i}"] = read_encoded_value(s)

        # From here, the mask is equal to the outlines
        cls.parse_outlines(s, maskdata)
        return dict(maskdata)



from __future__ import annotations

import logging

from io import BufferedReader
from typing import Any, Dict, Tuple
from vfbLib.parsers import read_encoded_value, uint8, uint16


logger = logging.getLogger(__name__)


class VfbHeaderParser:
    stream: BufferedReader | None = None

    @classmethod
    def parse(cls, stream: BufferedReader) -> Tuple[Dict[str, Any], int]:
        cls.stream = stream
        header: Dict[str, Any] = {}
        header["header0"] = cls.read_uint8()
        header["filetype"] = cls.stream.read(5).decode("cp1252")
        header["header1"] = cls.read_uint16()
        header["header2"] = cls.read_uint16()
        header["reserved"] = str(cls.stream.read(34))
        header["header3"] = cls.read_uint16()
        header["header4"] = cls.read_uint16()
        header["header5"] = cls.read_uint16()
        header["header6"] = cls.read_uint16()
        header["header7"] = cls.read_uint16()
        if header["header7"] == 10:
            # FL5 additions over FL3
            header["header8"] = cls.read_uint16()
            for i in range(9, 12):
                key = cls.read_uint8()
                val = read_encoded_value(stream)
                header[f"header{i}"] = {key: val}
            header["header12"] = cls.read_uint8()
            header["header13"] = cls.read_uint16()
        else:
            header["header13"] = header["header7"]
            del header["header7"]
        header["header14"] = cls.read_uint16()

        # Get the size of the original binary data
        datasize = cls.stream.tell()

        return header, datasize

    @classmethod
    def read_uint8(cls) -> int:
        assert cls.stream is not None
        return int.from_bytes(cls.stream.read(uint8), byteorder="little", signed=False)

    @classmethod
    def read_uint16(cls) -> int:
        assert cls.stream is not None
        return int.from_bytes(cls.stream.read(uint16), byteorder="little", signed=False)



from __future__ import annotations

from argparse import ArgumentParser
from fontTools.misc.textTools import deHexStr, hexStr
from io import BytesIO
from vfbLib.compilers.value import write_encoded_value
from vfbLib.parsers.value import read_encoded_value
from vfbLib.version import build_date


def yuri():
    parser = ArgumentParser(
        description=(f"vfbtool\nCopyright (c) 2023 by LucasFonts\nBuild {build_date}")
    )
    parser.add_argument(
        "-e",
        "--encode",
        action="store_true",
        default=False,
        help="Encode value instead of decoding",
    )
    parser.add_argument(
        "-l",
        "--long",
        action="store_true",
        default=False,
        help="Output longest notation when encoding",
    )
    parser.add_argument(
        "-s",
        "--signed",
        action="store_true",
        default=False,
        help="Treat value as signed",
    )
    parser.add_argument(
        "hexstring",
        type=str,
        nargs="+",
        help="Input hex string",
    )
    args = parser.parse_args()
    if args:
        if args.encode:
            stream = BytesIO()
            for value in args.hexstring:
                write_encoded_value(int(value), stream, args.signed)
            print(hexStr(stream.getvalue()))
        else:
            data = deHexStr("".join(args.hexstring))
            stream = BytesIO(data)
            while True:
                try:
                    print(read_encoded_value(stream, args.signed))
                except EOFError:
                    break

    else:
        parser.print_help()



from __future__ import annotations

import codecs
import json
import logging

from argparse import ArgumentParser
from difflib import unified_diff, HtmlDiff
from pathlib import Path
from vfbLib.vfb.vfb import Vfb
from vfbLib.version import build_date


logger = logging.getLogger(__name__)


def diffvfb():
    parser = ArgumentParser(
        description=(f"diffvfb\nCopyright (c) 2023 by LucasFonts\nBuild {build_date}")
    )
    parser.add_argument(
        "--html",
        type=str,
        help="Output diff in HTML format to file path",
    )
    parser.add_argument(
        "file1",
        type=str,
        nargs=1,
        help="First input file path (.vfb)",
    )
    parser.add_argument(
        "file2",
        type=str,
        nargs=1,
        help="Second input file path (.vfb)",
    )
    args = parser.parse_args()
    if not args:
        parser.print_help()
        return

    vfb1_path = Path(args.file1[0])
    vfb2_path = Path(args.file2[0])

    vfb1 = Vfb(vfb_path=vfb1_path, timing=False)
    vfb2 = Vfb(vfb_path=vfb2_path, timing=False)
    vfb1.read()
    vfb2.read()
    vfb1.decompile()
    vfb2.decompile()
    vfb1_str = json.dumps(vfb1.as_dict(), ensure_ascii=False, indent=4).splitlines()
    vfb2_str = json.dumps(vfb2.as_dict(), ensure_ascii=False, indent=4).splitlines()
    if args.html:
        html_diff = HtmlDiff()
        html = html_diff.make_file(
            vfb1_str, vfb2_str, str(vfb1_path), str(vfb2_path), context=True, numlines=5
        )
        with codecs.open(args.html[0], "wb", "utf-8") as f:
            f.write(html)
    else:
        d = unified_diff(vfb1_str, vfb2_str, str(vfb1_path), str(vfb2_path))
        for line in d:
            print(line)



from __future__ import annotations

from io import BytesIO
from struct import pack
from typing import Any, List
from vfbLib.compilers.value import write_encoded_value, write_value_5


# Compilers for VFB entries


class BaseCompiler:
    """
    Base class to compile vfb data.
    """

    def compile(self, data: Any, master_count: int = 0) -> bytes:
        """
        Compile the JSON-like main data structure and return the compiled binary data.

        Args:
            data (Any): The main data structure.
            master_count (int, optional): The number of masters. Defaults to 0.

        Returns:
            bytes: The compiled binary data.
        """
        self.master_count = master_count
        self.stream = BytesIO()
        self._compile(data)
        return self.stream.getvalue()

    def _compile(self, data: Any) -> None:
        raise NotImplementedError

    @classmethod
    def merge(cls, masters_data: List[Any], data: Any) -> None:
        """
        Merge the data of additional masters into the main data structure. This operates
        on the uncompiled JSON-like data structure.

        Args:
            masters_data (List[Any]): The additional masters data as a list with one
                entry per master.
            data (Any): The main data structure.
        """
        # Must be implemented for compilers that need it, e.g. the GlyphCompiler.
        pass

    def write_bytes(self, value: bytes) -> None:
        """
        Write binary data to the stream.

        Args:
            value (bytes): The data.
        """
        self.stream.write(value)

    def write_encoded_value(self, value: int, shortest=True) -> None:
        """
        Encode and write an int value to the stream. Optionally don't apply the length
        encoding optimization.

        Args:
            value (int): The value to write to the stream.
            shortest (bool, optional): Whether to write in the shortest possible
                notation. Defaults to True.
        """
        if shortest:
            write_encoded_value(value, self.stream)
        else:
            write_value_5(value, self.stream)

    def write_float(self, value: float, fmt: str = "d") -> None:
        """
        Write a float value to the stream.
        """
        encoded = pack(fmt, value)
        self.stream.write(encoded)

    def write_uint1(self, value: int) -> None:
        """
        Write a 1-byte unsigned value to the stream.
        """
        encoded = pack(">B", value)
        self.stream.write(encoded)

    def write_uint8(self, value: int) -> None:
        """
        Write a uint8 value to the stream.
        """
        encoded = pack(">H", value)
        self.stream.write(encoded)



from __future__ import annotations

import logging

from colorsys import hls_to_rgb
from typing import TYPE_CHECKING, Any, Dict, List, Tuple
from vfbLib.ufo.vfb2ufo import vfb2ufo_label_codes

if TYPE_CHECKING:
    from vfbLib.typing import (
        Anchor,
        GuidePropertyList,
        GuideDict,
        HintDict,
        LinkDict,
        MMNode,
    )


logger = logging.getLogger(__name__)


class VfbToUfoGlyph:
    def __init__(self) -> None:
        self.anchors: List[Anchor] = []
        self.guide_properties: GuidePropertyList = []
        self.hintmasks: List[Dict[str, int]] = []
        self.labels: Dict[str, int] = {}
        self.lib: Dict[str, Any] = {}
        self.links: LinkDict = {}
        self.mm_anchors: List[Any] | None = None
        self.mm_components: List[Any] = []
        self.mm_guides: GuideDict | None = None
        self.mm_hints: HintDict = {"h": [], "v": []}
        self.mm_metrics: List[Tuple[int, int]] = []
        self.mm_nodes: List[MMNode] = []
        self.name: str | None = None
        self.note: str | None = None
        self.point_labels: Dict[int, str] = {}
        self.rename_points: Dict[str, str]
        self.tth_commands: List[Dict[str, str | bool]] = []
        self.unicodes: List[int] = []

    def get_point_label(self, index: int, code: str, start_count: int = 1) -> str:
        if index in self.point_labels:
            # We already have a label for this point index
            return self.point_labels[index]

        # Special points
        num_nodes = len(self.mm_nodes)
        if index == num_nodes:
            return "lsb"
        elif index == num_nodes + 1:
            return "rsb"

        # Make a new label
        label_short = vfb2ufo_label_codes[code]
        i = start_count
        label = "%s%02d" % (label_short, i)
        while label in self.labels:
            i += 1
            label = "%s%02d" % (label_short, i)
        self.labels[label] = index
        self.point_labels[index] = label
        return label

    def set_mark(self, hue):
        self.lib["public.markColor"] = "%0.4f,%0.4f,%0.4f,1" % hls_to_rgb(
            h=hue / 255, l=0.8, s=0.76
        )
